/*
http://www.codeguru.com/cpp_mfc/EditDist.shtml
http://www.merriampark.com/ld.htm#CPLUSPLUS

  
                  
  
             CEditDist Abstract Template Class for Edit Distance Calculation on Generic Data Types 



             This article was contributed by Zvika Ben-Haim.

             Environment: Any modern C++ compiler

             Purpose

             The CEditDist class performs edit distance calculations on abstract data types.  The edit distance is defined as the
             minimum cost required to convert one string into another, where the conversion can include changing one character
             to another, deleting characters and inserting characters, with user-defined costs for each basic operation.  The
             algorithm is O(nm) where n and m are the lengths of the two strings compared (this is the fastest known
             algorithm).  Edit distance calculations are useful for finding the degree of similarity between strings, e.g. in
             non-exact database queries.

             The package is implemented as an abstract template base class.  The programmer derives from this base class in
             order to define the data types used and the cost functions.  The package includes complete documentation and an
             example of a simple implementation. 

             Example

             Overview: 
             This example will create a class that can be used to calculate the edit distance between two arrays of integers,
             according to the following cost function: Deleting or inserting an integer will have a cost of 5; Changing from one
             integer value to a different value will have a cost of 3. 

             Step 1: Instantiate the Base Class. 
             The instantiation must be explicitly performed because of the special treatment of abstract template classes in
             C++.  In our case, we want a base class of integers, so we will use the following statement:

             template class CEditDist<int>;

             Step 2: Derive from CEditDist. 
             We must now create a derivation from the base class.  This derivation will define the cost functions, which are
             virtual functions called from the base class during the edit distance calculation.

             class CIntEditClass : public CEditDist<int>
             {
             public:
                int DeleteCost(const int& deleted, int x, int y) { return 5; };
                int InsertCost(const int& inserted,int x, int y) { return 5; };
                int ChangeCost(const int& from, const int& to, int x, int y)
                      { return (from==to ? 0 : 3); };
             };

             Step 3: Using the Class. 
             To use the class, construct an object from the base class and call its EditDistance function, like so:

             CIntEditDist ed;
             int a[7] = {1,1,2,3,1,2,2};
             int b[6] = {1,2,2,3,1,1};
             int c = ed.EditDistance(a,7,b,6);

             In the above example, the edit distance stored in variable c should be 11.  This is because the least expensive edit
             replaces the second character in a from 1 to 2, replaces the sixth character from 2 to 1, and deletes the last
             character in a, for a total of two changes and one deletion, with a cost of 3+3+5=11. 
                  

                  

                                                        C++

In C++, the size of an array must be a constant, and this code fragment causes an error at compile time: 

int sz = 5;
int arr[sz];

This limitation makes the following C++ code slightly more complicated than it would be if the matrix could simply be declared as a two-dimensional
array, with a size determined at run-time. 

Here is the definition of the class (distance.h): 
*/

class Distance
{
  public:
    int LD (char const *s, char const *t);
  private:
    int Minimum (int a, int b, int c);
    int *GetCellPointer (int *pOrigin, int col, int row, int nCols);
    int GetAt (int *pOrigin, int col, int row, int nCols);
    void PutAt (int *pOrigin, int col, int row, int nCols, int x);
}; 

  /*
Here is the implementation of the class (distance.cpp): 


  */


// #include "distance.h"
#include <string.h>
#include <malloc.h>

//****************************
// Get minimum of three values
//****************************

int Distance::Minimum (int a, int b, int c)
{
int mi;

  mi = a;
  if (b < mi) {
    mi = b;
  }
  if (c < mi) {
    mi = c;
  }
  return mi;

}

//**************************************************
// Get a pointer to the specified cell of the matrix
//************************************************** 

int *Distance::GetCellPointer (int *pOrigin, int col, int row, int nCols)
{
  return pOrigin + col + (row * (nCols + 1));
}

//*****************************************************
// Get the contents of the specified cell in the matrix 
//*****************************************************

int Distance::GetAt (int *pOrigin, int col, int row, int nCols)
{
int *pCell;

  pCell = GetCellPointer (pOrigin, col, row, nCols);
  return *pCell;

}

//*******************************************************
// Fill the specified cell in the matrix with the value x
//*******************************************************

void Distance::PutAt (int *pOrigin, int col, int row, int nCols, int x)
{
int *pCell;

  pCell = GetCellPointer (pOrigin, col, row, nCols);
  *pCell = x;

}

//*****************************
// Compute Levenshtein distance
//*****************************

int Distance::LD (char const *s, char const *t)
{
int *d; // pointer to matrix
int n; // length of s
int m; // length of t
int i; // iterates through s
int j; // iterates through t
char s_i; // ith character of s
char t_j; // jth character of t
int cost; // cost
int result; // result
int cell; // contents of target cell
int above; // contents of cell immediately above
int left; // contents of cell immediately to left
int diag; // contents of cell immediately above and to left
int sz; // number of cells in matrix

  // Step 1     

  n = strlen (s);
  m = strlen (t);
  if (n == 0) {
    return m;
  }
  if (m == 0) {
    return n;
  }
  sz = (n+1) * (m+1) * sizeof (int);
  d = (int *) malloc (sz);

  // Step 2

  for (i = 0; i <= n; i++) {
    PutAt (d, i, 0, n, i);
  }

  for (j = 0; j <= m; j++) {
    PutAt (d, 0, j, n, j);
  }

  // Step 3

  for (i = 1; i <= n; i++) {

    s_i = s[i-1];

    // Step 4

    for (j = 1; j <= m; j++) {

      t_j = t[j-1];

      // Step 5

      if (s_i == t_j) {
        cost = 0;
      }
      else {
        cost = 1;
      }

      // Step 6 

      above = GetAt (d,i-1,j, n);
      left = GetAt (d,i, j-1, n);
      diag = GetAt (d, i-1,j-1, n);
      cell = Minimum (above + 1, left + 1, diag + cost);
      PutAt (d, i, j, n, cell);
    }
  }

  // Step 7

  result = GetAt (d, n, m, n);
  free (d);
  return result;
        
}